	title	'BIOS fuer FELTRON Compulady'
	page	68
;------------------------------------------------------
;Basic Input Output System fuer FELTRON Compulady
;
;Version 2.0 vom 20.Juni 1982
;
;Laufwerke A, B, C, D
;          double-density (512 Byte/Sektor)
;          double-side
;------------------------------------------------------

MSIZE	EQU	60	;MEMORY SIZE.

	MACLIB	Z80
	MACLIB	DISKDEF

TRUE	EQU	0FFFFH
FALSE	EQU	NOT TRUE

DBLSID	EQU	TRUE	;DRIVE-A

;------------------------------------------------------
; COMPULADY Disk monitor entry point
MON	EQU	0F800H
;
LSTS	EQU	MON+3
CSTS	EQU	MON+6
CI	EQU	MON+9
CO	EQU	MON+12
LO	EQU	MON+15
PO	EQU	MON+18
RI	EQU	MON+21
HOMEIT	EQU	MON+24
SETFLP	EQU	MON+27
SEEKIT	EQU	MON+30
SETREC	EQU	MON+33
DMA	EQU	MON+36
READIT	EQU	MON+39
WRITIT	EQU	MON+42
SETHD	EQU	MON+45
ONLINE	EQU	MON+48
SETCFG	EQU	MON+51
CONFG	EQU	MON+54
MULTIS	EQU	MON+57
MULTIT	EQU	MON+60
;
CBASE	EQU	(MSIZE-20)*1024	;BIAS FOR LARGER THAN 20K.
CPMB	EQU	CBASE+3400H	;START OF CPM.
BDOS	EQU	CPMB+806H	;START OF BDOS.
BIOS	EQU	CPMB+1600H	;BASE OF BIOS
CPML	EQU	BIOS-CPMB	;LENGTH OF CPM SYSTEM-BIOS.
NSECTS	EQU	CPML/128	;NUMBER OF SECTORS IN IT.
iobyt	equ	0003h
cdisk	equ	0004h
;
;------------------------------------------------------
	ORG	BIOS	;ORIGIN OF THIS PROGRAM
;
; I/O JUMP VECTOR
; THIS IS WHERE CPM CALLS WHENEVER IT NEEDS
; TO DO ANY INPUT/OUTPUT OPERATION.
;
	JMP	BOOT	;FROM COLD START LOADER.
WBOOTE:	JMP	WBOOT	;FROM WARM BOOT.
	JMP	CSTS	;CHECK CONSOLE KB STATUS.
conin:	JMP	CI	;READ CONSOLE CHARACTER.
conot:	JMP	co	;WRITE CONSOLE CHARACTER.
	JMP	LO	;WRITE LISTING CHAR.
	JMP	PO	;WRITE PUNCH CHAR.
	JMP	RI	;READ READER CHAR.
	JMP	HOME	;MOVE DISK TO TRACK ZERO.
	JMP	SELDSK	;SELECT DISK DRIVE.
	JMP	SETTRK	;SEEK TO TRACK IN REG A.
	JMP	SETSEC	;SET SECTOR NUMBER.
	JMP	SETDMA	;SET DISK STARTING ADR.
	JMP	READ	;READ SELECTED SECTOR.
	JMP	WRITE	;WRITE SELECTED SECTOR.
	JMP	LSTS	;RETURN LIST STATUS
	JMP	SECTRAN	;SECTOR TRANSLATE
;
;------------------------------------------------------
; BOOT
; THIS SECTION IS EXECUTED WHENEVER RESET AND RUN
; IS PUSHED, AFTER THE COLDSTART LOADER READS IN
; THE CPM SYSTEM.
;
BOOT:	LXI	SP,80H	;SET STACK POINTER.
;
;-----------------------------------------------------------
	call	init
	lxi	h,vram
	shld	vo+1
;-----------------------------------------------------------
;
	LXI	H,SMSG	;PRINT OPENING MESSAGE.
	CALL	PMSG
	XRA	A
	STA	DISKNO
GOCPM:	MVI	A,0C3H	;PUT JMP TO WBOOT
	STA	0	;ADR AT ZERO.
	LXI	H,WBOOTE
	SHLD	1
	STA	5
	LXI	H,BDOS	;PUT JUMP TO BDOS
	SHLD	6	;AT ADR 5,6,7.
	LXI	B,80H	;SET DEFAULT DMA ADR.
	CALL	SETDMA
	XRA	A	;FOR DEBLOCK
	STA	HSTACT	;HOST BUF ACTIVE
	STA	UNACNT	;CLR UNALLOC COUNT
	LDA	DISKNO	;GET DISK NUMBER TO
	MOV	C,A	;PASS TO CCP IN C.
	mvi	a,94h
	sta	iobyt
	JMP	CPMB	;JUMP TO CCP.
;
;------------------------------------------------------
; WARM-BOOT: READ ALL OF CPM BACK IN
; EXCEPT BIOS FROM DRIVE-A, THEN JUMP TO CCP.
;
WBOOT:	LXI	SP,80H	;SET STACK POINTER.
	MVI	C,0
	CALL	SETFLP
	CALL	HOMEIT
	LXI	B,3*256+2
	CALL	SETCFG
;
	MVI	C,1
	CALL	SEEKIT
	MVI	C,1
	CALL	SETREC
	MVI	C,11
	CALL	MULTIS
	LXI	B,CPMB
	CALL	DMA
	CALL	READIT
	ORA	A
	JNZ     BOOTERR
	MVI	C,1
	CALL	MULTIS
ALDON:	JMP	GOCPM
;
;
;
BOOTERR:
	LXI	H,BOOTMSG ;GET ADDRESS OF "?BOOT ".
	CALL	PMSG	;PRINT IT.
	CALL	CONIN	;READ A CHAR FROM CONSOLE.
	JR	WBOOT	;DO A WARM BOOT.
;
;
; SELECT DISK NUMBER ACCORDING TO REGISTER C.
;
SELDSK: PUSH	B
	CALL	SETFLP
	POP	B
	LXI	H,0
	RC
	MOV	A,C
	STA	DISKNO
	STA	SEKDSK	;FOR DEBLOCK (IF NO.>0)
	MOV	L,C	;L=DISK NUMBER 
	MVI	H,0	;HIGH ORDER ZERO
	DAD	H	;*2
	DAD	H	;*4
	DAD	H	;*8
	DAD	H	;*16 (SIZE OF EACH HEADER)
	LXI	D,DPBASE
	DAD	D	;HL=.DPBASE(DISKNO*16)
	RET	;RETURN FROM SELDSK.
;
HOME:	LXI	H,0
	SHLD	SEKTRK
	JMP	HOMEIT
;
; SET TRACK NUMBER TO WHATEVER IS IN REGISTER C.
; ALSO PERFORM MOVE TO THE CORRECT TRACK (SEEK).
;
SETTRK:
	SBCD	SEKTRK	;FOR DEBLOCK
	CALL	CHKDBLK
	RNZ
	JMP	SEEKIT
;
;CHECK DEBLOCKING
;IF DOUBLE DENSITY Z=0
;   SINGLE DENSITY Z=1
;
CHKDBLK:
	MVI	A,1
	ORA	A
	RET
;
; SET DISK SECTOR NUMBER.
;
SETSEC:
	MOV	A,C
	STA	SEKSEC	;FOR DEBLOCK
	CALL	CHKDBLK
	RNZ
	JMP	SETREC
;
; TRANSLATE THE SECTOR GIVEN BY BC USING THE
; TRANSLATE TABLE GIVEN BY DE
SECTRAN:
	CALL	CHKDBLK
	JRNZ	DBLSECT
	XCHG	;HL=.TRANS
	DAD	B	;HL=.TRANS(SECTOR)
	MOV	L,M	;L = TRANS(SECTOR)
	MVI	H,0	;HL= TRANS(SECTOR)
	RET		;WITH VALUE IN HL
DBLSECT:
	MOV	H,B	;FOR DEBLOCK
	MOV	L,C
	RET
;
; SET DISK DMA ADDRESS.
;
SETDMA:
	SBCD	DMAADR	;FOR DEBLOCK
	CALL	CHKDBLK
	RNZ
	JMP	DMA
;
;
;
; READ THE SECTOR AT SECT, FROM THE PRESENT TRACK.
; USE STARTING ADDRESS AT DMAADD.
;
READ:
	CALL	CHKDBLK
	JNZ	D$READ	;IF DEBLOCK REQ
	JMP	READIT
;
; WRITE THE SECTOR AT SECT, ON THE PRESENT TRACK.
; USE STARTING ADDRESS AT DMAADD.
;
WRITE:
	CALL	CHKDBLK
	JNZ	D$WRITE ;IF DEBLOCK REQ
	JMP	WRITIT
;
; MOVE THE HEAD TO THE TRACK IN REGISTER A.
; PRINT THE MESSAGE AT H&L UNTIL A ZERO.
;
PMSG:	MOV	A,M	;GET A CHARACTER.
	ORA	A	;IF IT'S ZERO,
	RZ	;RETURN.
	MOV	C,A	;OTHERWISE,
	CALL	CONOT	;PRINT IT.
	INX	H	;INCREMENT H&L,
	JR	PMSG	;AND GET ANOTHER.
;
; CBIOS MESSAGES
;
BOOTMSG:DB	'?Boot ',0
;
SMSG:	DB	1bh,30h,1bh,4ah
	db	'****************************************'
	db	'****************************************'
	db	'FELTRON Compulady '
	DB	MSIZE/10+'0',MSIZE MOD 10 + '0'
	DB	'K CP/M 2.2 <Serien-Nummer CP2-002-07492>'
	DB	0DH,0AH
	db	'Laufwerke A:, B:, C:, D: double-density, double-side'
	db	0dh,0ah
	db	'****************************************'
	db	'****************************************'
	db	0dh,0ah
	db	0

;------------------------------------------------------
 
TOPC	EQU 0F800H	;START ADDR OF REFRESH MEM
CLM	EQU 80
ROW	EQU 24

;CRTC HD46505R ADDR
CRTC	EQU	74H		;CRTC DATA REG
ADDR	EQU	75H		;CRTC ADDR REG
DISPTM	EQU	76H		;CRTC DISPTMG, BIT 0

VRAM:	PUSH	H
	PUSH	D
	PUSH	B
	CALL	CRT
	POP	B
	POP	D
	POP	H
	RET

CRT:
	MOV	A,C
	RES	7,A	;STRIP PARITY

	LXI	H,STAT
	BIT	0,M	;DISPF ON ?
	JNZ	DSPF 	;YES
	BIT	6,M
	JNZ	ESCXX	;MULTI LETTER ESC-SEQUENCE
	BIT	7,M
	JNZ	ESC1	;1-LETTER ESC

	CPI	0DH
	JRZ	CR
	CPI	0AH
	JRZ	LF
	CPI	08H
	JZ	BS
	CPI	09H
	JZ	TAB
	CPI	1BH	;ESC
	JRNZ	OTHER
	SETB	7,M
	RET

OTHER:	MOV	A,C
	CPI	' '
	RC
 
CHRA:	CALL	ADRS	;HL <- CSR+STAD+TOPC
	CALL	CWAIT
	MOV	M,C	;WRITE V-RAM
	CALL	RIGHT	;CSR + 1
	RNC		;CSR <= 1920
	CALL	SCRL
	LXI	D,CLM-1
	JR	CR1	;MOVE CSR TO FIRST CLM
 
;CLEAR LINE
CLRLN:	CALL	GETCLM
	LHLD	CSR
	ORA	A
	DSBC	D
	LXI	B,CLM
	CALL	CLRS

CR:	CALL	GETCLM	;DE <- CLM ADDR
CR1:	LHLD	CSR
	ORA	A
	DSBC	D	;CSR-CLMADDR
	JMP	WCSR

CRLF:	CALL	CR
LF:	CALL	DOWN
	RNC	;NO SCROLL UP REQUIRED
 
SCRL:	LXI	H,STAT
	BIT	2,M
	JNZ	CLRC	;IF SCROL OFF
	LXI	H,CLM*ROW
	LXI	B,CLM
	CALL	CLRS	;CLEAR NEXT 1-LINE
	LHLD	STAD	;ADJ STAD & CSR
	LXI	D,CLM
	DAD	D
	MOV	A,H
	ANI	07H	;2K RAPAROUND FACTOR
			;HL <- HL AND 7FFH
	MOV	H,A
	CALL	WSTAD
	LHLD	CSR
	JMP	WCSR

 
RIGHT:	LHLD	CSR
	INX	H	;CSR+1
	JR	DOWN1
 
DOWN:	LHLD	CSR
	LXI	D,CLM
	DAD	D	;CSR+CLM(80)
DOWN1:	XCHG		;DE <- CSR+CLM(80)
	LXI	H,ROW*CLM-1
	ORA	A
	DSBC	D	;(ROW*CLM-1)-(CSR+CLM) : 1919-CSR+80
	RC		;NO SCROLL DOWN
	XCHG
	JMP	WCSR

BS:
LEFT:	LHLD	CSR
	LXI	D,1	;LEFT 1
	JR	UP1
 
UP:	LHLD	CSR
	LXI	D,CLM
UP1:	ORA	A
	DSBC	D	;CSR-CLM OR CSR-1
	RC		;NO SCROLL UP
	JMP	WCSR
 
;CLEAR END OF SCREEN
CLRES:	LXI	H,CLM*ROW
	LDED	CSR
	JR	CLREL1
 
;CLEAR END OF LINE
CLREL:	CALL	GETCLM	;DE:CLMADDR HL:CLM
CLREL1: ORA	A
	DSBC	D
	MOV	B,H
	MOV	C,L
	LHLD	CSR
 
;CLEAR LINE
;SELECTIVE CLEAR V-RAM WITHOUT CURSOR MOVEMENT
; HL <- BEGINING OF CSR
; BC <- LEN
CLRS:	LDED	STAD
	DAD	D
	XCHG	;DE <- STAD+START OF CSR
CLRS1:	CALL	ADRS1	;HL <- DE+TOPC
	MVI	M,' '
	INX	D
	DCX	B
	MOV	A,B
	ORA	C
	JRNZ	CLRS1
	RET
 
;RANDAM CURSOR ADDRESSING
;ESC = X+32 Y+32 (32 ASCII BIAS)
;0 < X < 79 :	0 < Y < 23
;(0,0): HOME POSITION
;CSR=(Y-32)*80+(X-32)
RNDCSR:
	LXI	H,ESCBF+1
	MOV	A,M
	SUI	32	;GET X
	MOV	B,A	;B <- X
	INX	H
	MOV	A,M
	SUI	32	;GET Y
	MVI	H,0
	MOV	L,A	;HL <- Y
	MOV	C,B
	MOV	B,H	;BC <- X
	MOV	D,H
	MOV	E,L	;DE <- Y
	DAD	H	;*2
	DAD	H	;*4
	DAD	D	;*5
	DAD	H	;*10
	DAD	H	;*20
	DAD	H	;*40
	DAD	H	;*80
	DAD	B	;ADD X
	JMP	WCSR

;EVERY 8-COLMUN
TAB:	CALL	GETCLM
	MOV	A,E
TAB0:	SUI	8
	JRNC	TAB0
	LHLD	CSR
TAB1:	INX	H
	INR	A
	JRNZ	TAB1
	JMP	WCSR
 
SCRLON: LXI	H,STAT
	RES	2,M
	RET

SCRLOFF:LXI	H,STAT
	SETB	2,M
	RET
 

;INIT 46505R CRTC & CLR SCREEN
INIT:	LXI	H,IDATA
	MVI	B,0
INIT1:	MOV	A,B
	OUT	ADDR
	MOV	A,M
	OUT	CRTC
	INR	B
	INX	H
	BIT	4,B	;16 ?
	JRZ	INIT1
 
	XRA	A
	STA	STAT

CLRC:	LXI	H,TOPC	;CLEAR SCREEN
	LXI	D,2048
CLRC1:	CALL	CWAIT
	MVI	M,' '
	INX	H
	DCX	D
	MOV	A,D
	ORA	E
	JRNZ	CLRC1
	XCHG		;HL <- 0
	CALL	WSTAD
 
CHOME:	LXI	H,0	;CURSOR <- 0
 
WCSR:	SHLD	CSR	;0 <= CSR < ROW*CLM
	LDED	STAD
	DAD	D
	MVI	B,14	;R14
	JR	WRCRTC

WSTAD:	SHLD	STAD
	MVI	B,12	;R12

; SET START, CORSOR ADRR
; B	<- REG #
; 12: STRAT, 14: CURSOR
; HL <- 16 BIT DATA
WRCRTC: MOV	A,B	;B <- REG #
	OUT	ADDR
	MOV	A,H 
	OUT	CRTC	;HI BYTE
	INR	B
	MOV	A,B
	OUT	ADDR
	MOV	A,L
	OUT	CRTC	;LOW BYTE
	RET
 
;GET COLUMN ADDR ON DE
;DE <- CLM ADDR
;HL <- CLM (80)
GETCLM: LHLD	CSR
	LXI	D,CLM
	ORA	A
GETC2:	DSBC	D
	JRNC	GETC2	;REPEAT UNTIL CSR-CLM < 0
	DAD	D	;HL <- MOD(CLM,CSR)
	XCHG
	RET
 
;GET ACTUAL ADDR
;RET : HL <- STAD+CSR+TOPC
ADRS:	LHLD	CSR
	LDED	STAD
	DAD	D
	XCHG
;DE <- STAD+CSR
ADRS1:	MOV	A,D
	ANI	07H
	MOV	D,A
	LXI	H,TOPC
	DAD	D
	RET

CWAIT:	IN	DISPTM
	RRC
	JRC	CWAIT
	RET
 
IDATA:	DB	108	;R0
	DB	CLM	;R1=80
	DB	89	;R2
	DB	8	;R3
	DB	31	;R4
	DB	0	;R5
	DB	ROW	;R6=24
	DB	27	;R7
	DB	0	;R8
	DB	9	;R9
	DB	69H	;R10
	DB	9H	;R11
	DB	0	;R12
	DB	0	;R13
	DB	0	;R14
	DB	0	;R15

DSPFON: LXI	H,STAT
	SETB	0,M
	RET

DSPF:	PUSH	B
	CPI	0AH
	JRNZ	DSPF1	;NOT LF
	CALL	CRLF
	MVI	C,0AH
DSPF1:	CALL	CHRA	;DISPLAY IT
	LXI	H,STAT
	POP	B
	MOV	A,C
	BIT	1,M
	JRZ	DSPF3
	CPI	'Z'
	JRNZ	DSPF3
	RES	0,M	;ESC,Z	 DSPF OFF
DSPF2:	RES	1,M
	RET
DSPF3:	CPI	1BH	;ESC ?
	JRNZ	DSPF2	;NO
	SETB	1,M	;GET ESC
	RET

ESCXX:	LHLD	ESCP
	MOV	M,A
	INX	H
	SHLD	ESCP
	LXI	H,ESCLEN
	DCR	M
	RNZ

	LXI	H,STAT
	RES	6,M
	LDA	ESCBF
	LXI	H,ESCX
ESCX1:	CMP	M
	INX	H
	JRZ	ESCX2
	INX	H
	INX	H
	INX	H
	JR	ESCX1
ESCX2:	INX	H
	JR	FUND


ESC1:	RES	7,M	;RESET ESC-SEQ
	LXI	H,ESC1BR
	MVI	B,LESC1BR/3
ES1:	CMP	M
	INX	H
	JRZ	FUND
	INX	H
	INX	H
	DJNZ	ES1

	MVI	B,LESCX/4
ES4:	CMP	M
	INX	H
	JRZ	ES5
	INX	H
	INX	H
	INX	H
	DJNZ	ES4
	RET

ES5:	MOV	B,M	;LEN
	LXI	H,ESCLEN
	MOV	M,B	;LEN
	INX	H
	MOV	M,A	;1-ST LETTER
	INX	H
	SHLD	ESCP
	LXI	H,STAT
	SETB	6,M
	RET

FUND:	MOV	E,M
	INX	H
	MOV	D,M
	XCHG
	PCHL		;JUMP TO 1-LETTER ESC SEQUENCE


;BRANCH TABLE FOR 1-LETTER ESC SEQUENSE
ESC1BR: DB	'0'
	DW	CHOME
	DB	'1'
	DW	RIGHT
	DB	'2'
	DW	DOWN
	DB	'3'
	DW	LEFT
	DB	'4'
	DW	UP
	DB	'Y'	;DISPF ON
	DW	DSPFON
	DB	'J'	;CLEAR END OF SCREEN
	DW	CLRES
	DB	'K'	;CLEAR END OF LINE
	DW	CLREL
	DB	'L'	;CLEAR LINE
	DW	CLRLN
	DB	'C'	;CLEAR CHARA DISP
	DW	CLRC
	DB	'I'	;INITIALIZE
	DW	INIT
	DB	'T'	;SCROL OFF
	DW	SCRLOFF
	DB	'S'	;SCROL ON
	DW	SCRLON
LESC1BR EQU	$-ESC1BR

ESCX:	DB	'=',2
	DW	RNDCSR	;ESC = X Y (+32 OFF SET)
LESCX	EQU	$-ESCX

;-----------------------------------------------------------

;	SECTOR DEBLOCKING ALGORITHMS FOR CP/M 2.0	
;	UTILITY MACRO TO COMPUTE SECTOR MASK
SMASK	MACRO	HBLK
;;	COMPUTE LOG2(HBLK), RETURN @X AS RESULT
;;	(2 ** @X = HBLK ON RETURN)
@Y	SET	HBLK
@X	SET	0
;;	COUNT RIGHT SHIFTS OF @Y UNTIL = 1
	REPT	8
	IF	@Y = 1
	EXITM
	ENDIF
;;	@Y IS NOT 1, SHIFT RIGHT ONE POSITION
@Y	SET	@Y SHR 1
@X	SET	@X + 1
	ENDM
	ENDM
;
;	 CP/M TO HOST DISK CONSTANTS		
;
BLKSIZ	EQU	2048		;CP/M ALLOCATION SIZE
HSTSIZ	EQU	512		;HOST DISK SECTOR SIZE
HSTSPT	EQU	15		;HOST DISK SECTORS/TRK
HSTBLK	EQU	HSTSIZ/128	;CP/M SECTS/HOST BUFF
CPMSPT	EQU	HSTBLK * HSTSPT	;CP/M SECTORS/TRACK
SECMSK	EQU	HSTBLK-1	;SECTOR MASK
	SMASK	HSTBLK	;COMPUTE SECTOR MASK
SECSHF	EQU	@X	;LOG2(HSTBLK)
;
;	BDOS CONSTANTS ON ENTRY TO WRITE		
;
WRALL	EQU	0	;WRITE TO ALLOCATED
WRDIR	EQU	1	;WRITE TO DIRECTORY
WRUAL	EQU	2	;WRITE TO UNALLOCATED
;
;	THE BDOS ENTRY POINTS GIVEN BELOW SHOW THE	
;	CODE WHICH IS RELEVANT TO DEBLOCKING ONLY.	
;
	IF	DBLSID
OFFST	EQU	4
DKS	EQU	562
	ELSE
OFFST	EQU	2
DKS	EQU	281
	ENDIF
;
	DISKS	4
	DISKDEF 0,1,CPMSPT,0,BLKSIZ,DKS,128,128,OFFST
	DISKDEF	1,0
	DISKDEF 2,0
	DISKDEF 3,0
;
;	THE READ ENTRY POINT TAKES THE PLACE OF	
;	THE PREVIOUS BIOS DEFINTION FOR READ.	
;
D$READ:
;
;READ THE SELECTED CP/M SECTOR
;
	XRA	A
	STA	UNACNT
	MVI	A,1
	STA	READOP	;READ OPERATION
	STA	RSFLAG	;MUST READ DATA
	MVI	A,WRUAL
	STA	WRTYPE	;TREAT AS UNALLOC
	JR	RWOPER	;TO PERFORM THE READ
;
;	THE WRITE ENTRY POINT TAKES THE PLACE OF	
;	THE PREVIOUS BIOS DEFINTION FOR WRITE.	
;
D$WRITE:
	;WRITE THE SELECTED CP/M SECTOR
	XRA	A	;0 TO ACCUMULATOR
	STA	READOP	;NOT A READ OPERATION
	MOV	A,C	;WRITE TYPE IN C
	STA	WRTYPE
	CPI	WRUAL	;WRITE UNALLOCATED?
	JRNZ	CHKUNA	;CHECK FOR UNALLOC
;
;	WRITE TO UNALLOCATED, SET PARAMETERS
	MVI	A,BLKSIZ/128	;NEXT UNALLOC RECS
	STA	UNACNT
	LDA	SEKDSK	;DISK TO SEEK
	STA	UNADSK	;UNADSK = SEKDSK
	LHLD	SEKTRK
	SHLD	UNATRK	;UNATRK = SECTRK
	LDA	SEKSEC
	STA	UNASEC	;UNASEC = SEKSEC
;
CHKUNA:
	;CHECK FOR WRITE TO UNALLOCATED SECTOR
	LDA	UNACNT	;ANY UNALLOC REMAIN?
	ORA	A
	JRZ	ALLOC	;SKIP IF NOT
;
;	MORE UNALLOCATED RECORDS REMAIN
	DCR	A	;UNACNT = UNACNT-1
	STA	UNACNT
	LDA	SEKDSK	;SAME DISK?
	LXI	H,UNADSK
	CMP	M	;SEKDSK = UNADSK?
	JRNZ	ALLOC	;SKIP IF NOT
;
;	DISKS ARE THE SAME
	LXI	H,UNATRK
	CALL	SEKTRKCMP	;SEKTRK = UNATRK?
	JRNZ	ALLOC	;SKIP IF NOT
;
;	TRACKS ARE THE SAME
	LDA	SEKSEC	;SAME SECTOR?
	LXI	H,UNASEC
	CMP	M	;SEKSEC = UNASEC?
	JRNZ	ALLOC	;SKIP IF NOT
;
;	MATCH, MOVE TO NEXT SECTOR FOR FUTURE REF
	INR	M	;UNASEC = UNASEC+1
	MOV	A,M	;END OF TRACK?
	CPI	CPMSPT	;COUNT CP/M SECTORS
	JRC	NOOVF	;SKIP IF NO OVERFLOW
;
;	OVERFLOW TO NEXT TRACK
	MVI	M,0	;UNASEC = 0
	LHLD	UNATRK
	INX	H
	SHLD	UNATRK	;UNATRK = UNATRK+1
;
NOOVF:
	;MATCH FOUND, MARK AS UNNECESSARY READ
	XRA	A	;0 TO ACCUMULATOR
	STA	RSFLAG	;RSFLAG=0
	JR	RWOPER	;TO PERFORM THE WRITE
ALLOC:
	;NOT AN UNALLOCATED RECORD, REQUIRES PRE-READ
	XRA	A	;0 TO ACCUM
	STA	UNACNT	;UNACNT = 0
	INR	A	;1 TO ACCUM
	STA	RSFLAG	;RSFLAG = 1
;
;	COMMON CODE FOR READ AND WRITE FOLLOWS	
;
RWOPER:
	;ENTER HERE TO PERFORM THE READ/WRITE
	XRA	A	;RERO TO ACCUM
	STA	ERFLAG	;NO ERRORS (YET)
	LDA	SEKSEC	;COMPUTE HOST SECTOR
	REPT	SECSHF
	ORA	A	;CARRY = 0
	RAR		;SHIFT RIGHT
	ENDM
	STA	SEKHST	;HOST SECTOR TO SEEK
;
;	ACTIVE HOST SECTOR?
	LXI	H,HSTACT	;HOST ACTIVE FLAG
	MOV	A,M
	MVI	M,1	;ALWAYS BECOMES 1
	ORA	A	;WAS IT ALREADY?
	JRZ	FILHST	;FILL HOST IF NOT
;
;	HOST BUFFER ACTIVE, SAME AS SEEK BUFFER?
	LDA	SEKDSK
	LXI	H,HSTDSK	;SAME DISK?
	CMP	M	;SEKDSK = HSTDSK?
	JRNZ	NOMATCH
;
;	SAME DISK, SAME TRACK?
	LXI	H,HSTTRK
	CALL	SEKTRKCMP	;SEKTRK = HSTTRK?
	JRNZ	NOMATCH
;
;	SAME DISK, SAME TRACK, SAME BUFFER?
	LDA	SEKHST
	LXI	H,HSTSEC	;SEKHST = HSTSEC?
	CMP	M
	JRZ	MATCH	;SKIP IF MATCH
;
NOMATCH:
	;PROPER DISK, BUT NOT CORRECT SECTOR
	LDA	HSTWRT	;HOST WRITTEN?
	ORA	A
	CNZ	WRITEHST	;CLEAR HOST BUFF
;
FILHST:
	;MAY HAVE TO FILL THE HOST BUFFER
	LDA	SEKDSK
	STA	HSTDSK
	LHLD	SEKTRK
	SHLD	HSTTRK
	LDA	SEKHST
	STA	HSTSEC
	LDA	RSFLAG	;NEED TO READ?
	ORA	A
	CNZ	READHST	;YES, IF 1
	XRA	A	;0 TO ACCUM
	STA	HSTWRT	;NO PENDING WRITE
;
MATCH:
	;COPY DATA TO OR FROM BUFFER
	LDA	SEKSEC	;MASK BUFFER NUMBER
	ANI	SECMSK	;LEAST SIGNIF BITS
	MOV	L,A	;READY TO SHIFT
	MVI	H,0	;DOUBLE COUNT
	REPT	7	;SHIFT LEFT 7
	DAD	H
	ENDM
;	HL HAS RELATIVE HOST BUFFER ADDRESS
	LXI	D,HSTBUF
	DAD	D	;HL = HOST ADDRESS
	LDED	DMAADR	;GET/PUT CP/M DATA
	LXI	B,128	;LENGTH OF MOVE
	LDA	READOP	;WHICH WAY?
	ORA	A
	JRNZ	RWMOVE	;SKIP IF READ
;
;	WRITE OPERATION, MARK AND SWITCH DIRECTION
	MVI	A,1
	STA	HSTWRT	;HSTWRT = 1
	XCHG		;SOURCE/DEST SWAP
;
RWMOVE: ;BC INITIALLY 128, HL IS SOURCE, DE IS DEST
	LDIR
;
;	DATA HAS BEEN MOVED TO/FROM HOST BUFFER
	LDA	WRTYPE	;WRITE TYPE
	CPI	WRDIR	;TO DIRECTORY?
	LDA	ERFLAG	;IN CASE OF ERRORS
	RNZ		;NO FURTHER PROCESSING
;
;	CLEAR HOST BUFFER FOR DIRECTORY WRITE
	ORA	A	;ERRORS?
	RNZ		;SKIP IF SO
	XRA	A	;0 TO ACCUM
	STA	HSTWRT	;BUFFER WRITTEN
	CALL	WRITEHST
	LDA	ERFLAG
	RET
;
;	UTILITY SUBROUTINE FOR 16-BIT COMPARE	
;
SEKTRKCMP:
	;HL = .UNATRK OR .HSTTRK, COMPARE WITH SEKTRK
	XCHG
	LXI	H,SEKTRK
	LDAX	D	;LOW BYTE COMPARE
	CMP	M	;SAME?
	RNZ		;RETURN IF NOT
;	LOW BYTES EQUAL, TEST HIGH 1S
	INX	D
	INX	H
	LDAX	D
	CMP	M	;SETS FLAGS
	RET
;
;	WRITEHST PERFORMS THE PHYSICAL WRITE TO	
;	THE HOST DISK, READHST READS THE PHYSICAL	
;	DISK.
;
WRITEHST:
	;HSTDSK = HOST DISK #, HSTTRK = HOST TRACK #,
	;HSTSEC = HOST SECT #. WRITE "HSTSIZ" BYTES
	;FROM HSTBUF AND ZETURN ERROR FLAG IN ERFLAG.
	;RETURN ERFLAG NON-ZERO IF ERROR
	;
	CALL	RWHST
	CALL	WRITIT
	JR	RWHSTE
;
RWHST:
	LDA	HSTDSK
	MOV	C,A
	CALL	SETFLP
	LXI	B,HSTBUF
	CALL	DMA
	;
	LXI	H,XLT13
	LDA	HSTSEC
	MOV	E,A
	MVI	D,0
	DAD	D
	MOV	C,M
	CALL	SETREC
	;
	LBCD	HSTTRK
	IF	DBLSID
	MOV	B,C
	SRLR	B
	BIT	0,C
	JRNZ	HD1
	MVI	C,0
	JR	SEEK
HD1:	MVI	C,1
SEEK:	CALL	SETHD
	MOV	C,B
	ENDIF
	JMP	SEEKIT
;
READHST:
	;HSTDSK = HOST DISK #, HSTTRK = HOST TRACK #,
	;HSTSEC = HOST SECT #. READ "HSTSIR" BYTES
	;INTO HSTBUF AND RETURN ERROR FLAG IN ERFLAG.
	;
	CALL	RWHST
	CALL	READIT
RWHSTE: STA	ERFLAG
	RET
;
XLT13:	DB	1,6,11,2,7,12,3,8,13,4,9,14,5,10,15
;
AAAAAA: ;END OF CODE
;
;STD BIOS WORKING
DISKNO: DS	1
;
;-----------------------------------------------------------
stad:	ds	2
csr:	ds	2
stat:	ds	1
escp:	ds	2
esclen:	ds	1
escbf:	ds	20
;-----------------------------------------------------------

;	UNITIALIZED RAM DATA AREAS			
CONOTF:	DS	1
;
SEKDSK:	DS	1	;SEEK DISK NUMBER
SEKTRK:	DS	2	;SEEK TRACK NUMBER
SEKSEC:	DS	1	;SEEK SECTOR NUMBER
;
HSTDSK:	DS	1	;HGST DISK NUMBER
HSTTRK:	DS	2	;HOST TRACK NUMBER
HSTSEC:	DS	1	;HOST SECTOR NUMBER
;
SEKHST:	DS	1	;SEEK SHR SECSHF
HSTACT:	DS	1	;HOST ACTIVE FLAG
HSTWRT:	DS	1	;HOST WRITTEN FLAG
;
UNACNT:	DS	1	;UNALLOC REC CNT
UNADSK:	DS	1	;LAST UNALLOC DISK
UNATRK:	DS	2	;LAST UNALLOC TRACK
UNASEC:	DS	1	;LAST UNALLOC SECTOR
;
ERFLAG:	DS	1	;ERROR REPORTING
RSFLAG:	DS	1	;READ SECTOR FLAG
READOP:	DS	1	;1 IF READ OPERATION
WRTYPE:	DS	1	;WRITE OPERATION TYPE
DMAADR:	DS	2	;LAST DMA ADDRESS
HSTBUF:	DS	HSTSIZ	;HOST BUFFER
;
SIO1:	DS	2
SIO2:	DS	2
;	THE ENDEF MACRO INVOCATION GOES HERE		
;
	ENDEF
;
;
;-----------------------------------------------------------
	org	0f700h
;
kin:	ds	3
vo:	ds	3
atogl:	ds	1	
;-----------------------------------------------------------

	END
